
const f32 MAX_BUILD_LENGTH = 4.0f;

shared class BlockCursor
{
    Vec2f tileAimPos;
    bool cursorClose;
    bool buildable;
    bool supported;
    bool hasReqs;
};

void AddCursor( CBlob@ this )
{
    if (!this.exists("blockCursor"))
    {
        BlockCursor bc;
        bc.buildable = bc.cursorClose = false;
        this.set( "blockCursor", bc );
    }
}

bool isTileSolidOrPlatform( CMap@ map, TileType t )
{
	return map.isTileSolid( t ) || map.isTilePlatform( t );
}

bool isBuildableAtPos( CBlob@ this, Vec2f p, TileType buildTile, CBlob @blob )
{
    f32 radius = 0.0f;
    CMap@ map = this.getMap();

    if (blob is null) // BLOCKS
    {
        radius = map.tilesize;
    }
    else // BLOB
    {
        radius = blob.getRadius();
    }

//printf("a");
    // tilemap check
    bool buildSolid = (map.isTileSolid(buildTile) || (blob !is null && blob.isCollidable()));
    Vec2f tilespace = map.getTileSpacePosition(p);
    int offset = map.getTileOffsetFromTileSpace( tilespace );
    TileType backtile = map.getTile(offset);
    TileType left = map.getTile(offset-1);
    TileType right = map.getTile(offset+1);
    TileType up = map.getTile(offset-map.tilemapwidth);
    TileType down = map.getTile(offset+map.tilemapwidth);

    if (buildTile > 0 && buildTile < 255 && blob is null && buildTile == backtile) {
        return false;
    }

//printf("b");
	if (backtile == CMap::tile_wood && buildTile == CMap::tile_castle) { // can build stone on wood
	}
	else
    if (map.isTileSolid(backtile) && (buildSolid || (!buildSolid && !map.isTileBlob(backtile)))) { // don't build on solid unless we are building background and placing behind a blob (Door)
        return false;
    }

//printf("c");
    bool canPlaceOnBackground = ((blob is null) || (blob.getShape().getConsts().support > 0));   // if this is a blob it has to do support - so spikes cant be placed on back

    if (
        (!canPlaceOnBackground || !map.isTileBackgroundNonEmpty(backtile)) &&      // can put against background
        !(                                              // can put sticking next to something
            isTileSolidOrPlatform( map, left ) || (canPlaceOnBackground && map.isTileBackgroundNonEmpty( left ))  ||
            isTileSolidOrPlatform( map, right) || (canPlaceOnBackground && map.isTileBackgroundNonEmpty( right )) ||
            isTileSolidOrPlatform( map, up )   || (canPlaceOnBackground && map.isTileBackgroundNonEmpty( up ))    ||
            isTileSolidOrPlatform( map, down ) || (canPlaceOnBackground && map.isTileBackgroundNonEmpty( down ))
        )
    )
    {
        return false;
    }
    // no blocking actors?

	if (blob is null || !blob.hasTag("ignore blocking actors"))
	{

		Vec2f middle = p;
		if (blob is null)
			middle += Vec2f(map.tilesize*0.5f, map.tilesize*0.5f);

		if (buildSolid && map.getSectorAtPosition( middle, "no build") !is null) {
			return false;
		}

		const string name = blob !is null ? blob.getName() : "";
		CBlob@[] blobsInRadius;	   
		if (map.getBlobsInRadius( middle, buildSolid ? map.tilesize : 0.0f, @blobsInRadius )) 
		{
			for (uint i = 0; i < blobsInRadius.length; i++)
			{
				CBlob @b = blobsInRadius[i];
				if (buildSolid)
				{	 				
					if (b is this)					// this is me
					{
						if ( (b.getPosition() - middle).getLength() < map.tilesize*0.8f)
							return false;
					}
					else 
					{
						bool player = b.hasTag("player");
						f32 radiusMod = player ? 1.5f : 0.5f;
						if (b !is blob && ((b.getPosition() - middle).getLength() <= b.getRadius()*radiusMod))				// cant place on any other blob
							return false;
					}
				}
				else // no stacking of non-collidable objects (spikes)
				{  									
					if (b !is this)
					{		
						if (name == b.getName())		
							return false;	
					}	
				}
			}
		}
	}

	return true;
}

void SetTileAimpos( CBlob@ this, BlockCursor@ bc )
{
    // calculate tile mouse pos
    Vec2f pos = this.getPosition();
    Vec2f aimpos = this.getAimPos();
    Vec2f mouseNorm = aimpos - pos;
    f32 mouseLen = mouseNorm.Length();
    const f32 maxLen = MAX_BUILD_LENGTH;
    mouseNorm /= mouseLen;

    if (mouseLen > maxLen * getMap().tilesize)
    {
        f32 d = maxLen * getMap().tilesize;
        Vec2f p = pos + Vec2f( d * mouseNorm.x, d * mouseNorm.y);
        p = getMap().getTileSpacePosition( p );
        bc.tileAimPos = getMap().getTileWorldPosition( p );
    }
    else
    {
        bc.tileAimPos = getMap().getTileSpacePosition( aimpos );
        bc.tileAimPos = getMap().getTileWorldPosition( bc.tileAimPos );
    }

    bc.cursorClose = (mouseLen < (maxLen + 0.51f) * getMap().tilesize);
}

void SetupBuildDelay( CBlob@ this )
{
    this.set_u32("build time", getGameTime() );
    this.set_u32("build delay", 17 ); // move this to builder init
}

bool isBuildDelayed( CBlob@ this )
{
    return (getGameTime() <= this.get_u32("build time") );
}

void SetBuildDelay( CBlob@ this )
{
    SetBuildDelay( this, this.get_u32("build delay") );
}

void SetBuildDelay( CBlob@ this, uint time )
{
	this.set_u32("build time", getGameTime() + time );
}

bool isBuildRayBlocked( Vec2f pos, Vec2f target )
{
    CMap@ map = getMap();

	Vec2f vector = target - pos;
	vector.Normalize();
	target -= vector * map.tilesize;
    
	f32 halfsize = map.tilesize * 0.5f;
    
    f32 yoff = 0.0f;
    f32 xoff = 0.0f;
    
    f32 ydiff = target.y - pos.y;
    if (Maths::Abs(ydiff) > halfsize)
    {
		yoff = halfsize;
		//if (ydiff < 0)
			yoff *= -1.0f;
	}
	
	f32 xdiff = target.x - pos.x;
    if (Maths::Abs(xdiff) > halfsize)
    {
		xoff = halfsize;
		if (xdiff < 0)
			xoff *= -1.0f;
	}
		
	return map.rayCastSolid( pos + Vec2f(xoff, yoff), target );
}
