//Explode.as - Explosions

/**
 *
 * used mainly for void Explode ( CBlob@ this, f32 radius, f32 damage )
 *
 * the effect of the explosion can be customised with properties:
 *
 * f32 map_damage_radius        - the radius to damage the map in
 * f32 map_damage_ratio         - the ratio of part-damage to full-damage of the map
 *                                  0.0 is all part-damage, 1.0 is all full-damage
 * bool map_damage_raycast      - whether to damage through terrain, or just the surface blocks;
 *
 * string custom_explosion_sound - the sound played when the explosion happens
 *
 * u8 custom_hitter             - the hitter from Hitters.as to use
 */


#include "Hitters.as";
#include "ShieldCommon.as";

bool isOwnerBlob(CBlob@ this, CBlob@ that)
{
    if (!this.exists("explosive_parent")) { return false; }

    return (that.getNetworkID() == this.get_u16("explosive_parent"));
}

void makeSmallExplosionParticle(Vec2f pos)
{
    ParticleAnimated( "Entities/Effects/Sprites/SmallExplosion"+(Random(3)+1)+".png",
                      pos, Vec2f(0,0.5f), 0.0f, 1.0f,
                      3+Random(3),
                      -0.1f, true );
}

void Explode ( CBlob@ this, f32 radius, f32 damage )
{
    Vec2f pos = this.getPosition();
    CMap@ map = getMap();

    if (!this.exists("custom_explosion_sound")) {
        Sound::Play( "Bomb.ogg", this.getPosition() );
    }
    else {
        Sound::Play( this.get_string("custom_explosion_sound"), this.getPosition() );
    }

    ParticleAnimated( "Entities/Effects/Sprites/Explosion.png", pos, Vec2f(0,0.5f), 0.0f, 1.0f, 3+Random(3), -0.1f, true );
    //load custom properties
    //map damage
    f32 map_damage_radius = 0.0f;

    if (this.exists("map_damage_radius")) {
        map_damage_radius = this.get_f32("map_damage_radius");
    }

    f32 map_damage_ratio = 0.5f;

    if (this.exists("map_damage_ratio")) {
        map_damage_ratio = this.get_f32("map_damage_ratio");
    }

    bool map_damage_raycast = true;

    if (this.exists("map_damage_raycast")) {
        map_damage_raycast = this.get_bool("map_damage_raycast");
    }

    //actor damage
    u8 hitter = Hitters::explosion;

    if (this.exists("custom_hitter")) {
        hitter = this.get_u8("custom_hitter");
    }

    const int r = (radius * (2.0/3.0));

    for (int i = 0; i < radius * 0.16; i++)
    {
        Vec2f partpos = pos + Vec2f( Random(r*2)-r, Random(r*2)-r );
        Vec2f endpos = partpos;

        if (map !is null)
        {
            if (!map.rayCastSolid(pos, partpos, endpos))
				makeSmallExplosionParticle(endpos);
        }        
    }

    if (getNet().isServer())
    {
        //hit map if we're meant to
        if (map_damage_radius > 0.1f)
        {
            int tile_rad = int( map_damage_radius / map.tilesize ) + 1;
            f32 rad_thresh = map_damage_radius*map_damage_ratio;
            Vec2f m_pos = (pos / map.tilesize);
            m_pos.x = Maths::Floor(m_pos.x);
            m_pos.y = Maths::Floor(m_pos.y);
            m_pos = (m_pos * map.tilesize) + Vec2f(map.tilesize / 2, map.tilesize / 2);

            //explode outwards
            for (int x_step = 0; x_step <= tile_rad; ++x_step)
            {
                for (int y_step = 0; y_step <= tile_rad; ++y_step)
                {
                    Vec2f offset = (Vec2f(x_step, y_step) * map.tilesize);

                    for (int i = 0; i < 4; i++)
                    {
                        if (i == 1)
                        {
                            if (x_step == 0) { continue; }

                            offset.x = -offset.x;
                        }

                        if (i == 2)
                        {
                            if (y_step == 0) { continue; }

                            offset.y = -offset.y;
                        }

                        if (i == 3)
                        {
                            if (x_step == 0) { continue; }

                            offset.x = -offset.x;
                        }

                        f32 dist = offset.Length();

                        if (dist < map_damage_radius)
                        {
                            //do we need to raycast?
                            bool canHit = !map_damage_raycast || (dist < 0.1f);

                            if (!canHit)
                            {
                                Vec2f v = offset;
                                v.Normalize();
                                v = v * (dist - map.tilesize);
                                canHit = ! ( map.rayCastSolid (m_pos, m_pos + v) );
								if (map.getSectorAtPosition( m_pos + v, "no build") !is null) {
									canHit = false;
								}
                            }

                            if (canHit)
                            {
                                if (dist > rad_thresh)
                                {
                                    map.server_DestroyTile(m_pos + offset, 1.0f);
                                }
                                else
                                {
                                    map.server_DestroyTile(m_pos + offset, 100.0f);
                                }
                            }
                        }
                    }
                }
            }

            //end loops
        }

        //hit blobs
        CBlob@[] blobs;
        map.getBlobsInRadius( pos, radius, @blobs );

        for (uint i = 0; i < blobs.length; i++)
        {
            CBlob@ hit_blob = blobs[i];
			if (hit_blob is this)
				continue;

            Vec2f hit_blob_pos = hit_blob.getPosition();
            Vec2f wall_hit;

			// no wall in front

            if (map.rayCastSolidNoBlobs(pos, hit_blob_pos, wall_hit)) { continue; } 

			// no blobs in front

			HitInfo@[] hitInfos; 
			Vec2f hitvec = hit_blob_pos - pos;		
			if (map.getHitInfosFromRay( pos, -hitvec.getAngle(), hitvec.getLength(), this, @hitInfos ))
			{
				bool blocked = false;
				for (uint i = 0; i < hitInfos.length; i++)
				{
					HitInfo@ hi = hitInfos[i];

					if (hi.blob !is null) // blob
					{	
						if (hi.blob is this || hi.blob is hit_blob || !hi.blob.isCollidable()) { 
							continue;
						}					   

						// only shield and heavy/medium things block explosions
						if (hi.blob.hasTag("medium weight") || hi.blob.hasTag("heavy weight") || hi.blob.getMass() > 500 || hi.blob.getShape().isStatic() || ( hi.blob.hasTag("shielded") && blockAttack( hi.blob, hitvec, 0.0f)))
						{
							blocked = true;
							break;
						}
					}
				}

				if (blocked)
					continue;
			}




            //get the offset and distance
            Vec2f offset = hit_blob_pos - pos;
            f32 distance = offset.Length();
            //set the scale (2 step)
            f32 scale = (distance > (radius* 0.7)) ? 0.5f : 1.0f;
            f32 dam = damage * scale;
            //explosion particle
            makeSmallExplosionParticle(hit_blob_pos);
            //the force, copy across
            Vec2f bombforce = offset;
            bombforce.Normalize();
            bombforce *= 2.0f;
            bombforce.y -= 0.2f; // push up for greater cinematic effect
            bombforce.x = Maths::Round(bombforce.x);
            bombforce.y = Maths::Round(bombforce.y);
            bombforce /= 2.0f;
            bombforce *= hit_blob.getMass() * (3.0f) * scale;

            this.server_Hit( hit_blob, hit_blob_pos, bombforce, dam, hitter, isOwnerBlob(this,hit_blob) || hit_blob.hasTag("dead")); //allow selfkill with bombs
        }
    }

    this.server_Die();
}

void onHitBlob( CBlob@ this, Vec2f worldPoint, Vec2f velocity, f32 damage, CBlob@ hitBlob, u8 customData )
{
    if (customData == Hitters::bomb)
    {
        hitBlob.AddForce( velocity );
    }
}	  
